<?php
/**
 * WP_Framework_Core Classes Main
 *
 * @author Technote
 * @copyright Technote All Rights Reserved
 * @license http://www.opensource.org/licenses/gpl-2.0.php GNU General Public License, version 2
 * @link https://technote.space
 */

namespace WP_Framework_Core\Classes;

use Exception;
use OutOfRangeException;
use WP_Error;
use WP_Framework;
use WP_Framework\Package_Base;
use WP_Framework_Admin\Classes\Models\Admin;
use WP_Framework_Api\Classes\Models\Api;
use WP_Framework_Cache\Classes\Models\Cache;
use WP_Framework_Common\Classes\Models\Array_Utility;
use WP_Framework_Common\Classes\Models\Config;
use WP_Framework_Common\Classes\Models\Define;
use WP_Framework_Common\Classes\Models\Deprecated;
use WP_Framework_Common\Classes\Models\File_Utility;
use WP_Framework_Common\Classes\Models\Filter;
use WP_Framework_Common\Classes\Models\Input;
use WP_Framework_Common\Classes\Models\Option;
use WP_Framework_Common\Classes\Models\Setting;
use WP_Framework_Common\Classes\Models\String_Utility;
use WP_Framework_Common\Classes\Models\System;
use WP_Framework_Common\Classes\Models\Uninstall;
use WP_Framework_Common\Classes\Models\User;
use WP_Framework_Common\Classes\Models\Utility;
use WP_Framework_Core\Interfaces\Package;
use WP_Framework_Core\Interfaces\Singleton;
use WP_Framework_Cron\Classes\Models\Cron;
use WP_Framework_Custom_Post\Classes\Models\Custom_Post;
use WP_Framework_Db\Classes\Models\Db;
use WP_Framework_Device\Classes\Models\Device;
use WP_Framework_Editor\Classes\Models\Editor;
use WP_Framework_Log\Classes\Models\Log;
use WP_Framework_Mail\Classes\Models\Mail;
use WP_Framework_Post\Classes\Models\Post;
use WP_Framework_Presenter\Classes\Models\Drawer;
use WP_Framework_Presenter\Classes\Models\Minify;
use WP_Framework_Session\Classes\Models\Session;
use WP_Framework_Social\Classes\Models\Social;
use WP_Framework_Test\Classes\Models\Test;
use WP_Framework_Update\Classes\Models\Update;
use WP_Framework_Update_Check\Classes\Models\Update_Check;
use WP_Framework_Upgrade\Classes\Models\Upgrade;

if ( ! defined( 'WP_CONTENT_FRAMEWORK' ) ) {
	exit;
}

/**
 * Class Main
 * @package WP_Framework_Core\Classes
 * @property Define $define
 * @property Config $config
 * @property Setting $setting
 * @property Filter $filter
 * @property Uninstall $uninstall
 * @property Utility $utility
 * @property Array_Utility $array
 * @property String_Utility $string
 * @property File_Utility $file
 * @property Option $option
 * @property User $user
 * @property Input $input
 * @property Deprecated $deprecated
 * @property System $system
 * @property Db $db
 * @property Log $log
 * @property Admin $admin
 * @property Api $api
 * @property Drawer $drawer
 * @property Minify $minify
 * @property Mail $mail
 * @property Test $test
 * @property Cron $cron
 * @property Custom_Post $custom_post
 * @property Device $device
 * @property Editor $editor
 * @property Session $session
 * @property Social $social
 * @property Post $post
 * @property Update $update
 * @property Update_Check $update_check
 * @property Upgrade $upgrade
 * @property Cache $cache
 */
class Main {

	/**
	 * @var Main[] $instances
	 */
	private static $instances = [];

	/**
	 * @var array $shared_object
	 */
	private static $shared_object = [];

	/**
	 * @var WP_Framework $app
	 */
	protected $app;

	/**
	 * @var bool $initialized
	 */
	private $initialized = false;

	/**
	 * @var array $properties
	 */
	private $properties;

	/**
	 * @var array $class_map
	 */
	private $class_map;

	/**
	 * @var array $class_target_package
	 */
	private $class_target_package;

	/**
	 * @var array $namespace_target_package
	 */
	private $namespace_target_package;

	/**
	 * @var WP_Framework|null[] $alternative_instances
	 */
	private $alternative_instances;

	/**
	 * @var array $property_instances
	 */
	private $property_instances = [];

	/**
	 * @param WP_Framework $app
	 *
	 * @return Main
	 */
	public static function get_instance( WP_Framework $app ) {
		if ( ! array_key_exists( $app->plugin_name, self::$instances ) ) {
			self::$instances[ $app->plugin_name ] = new self( $app );
		}

		return self::$instances[ $app->plugin_name ];
	}

	/**
	 * Main constructor.
	 *
	 * @param WP_Framework $app
	 */
	private function __construct( WP_Framework $app ) {
		$this->app = $app;
		$this->initialize();
	}

	/**
	 * @param string $name
	 *
	 * @return Singleton
	 * @throws OutOfRangeException
	 */
	public function __get( $name ) {
		return $this->get( $name );
	}

	/**
	 * @param string $name
	 *
	 * @return bool
	 */
	public function __isset( $name ) {
		return array_key_exists( $name, $this->properties );
	}

	/**
	 * initialize
	 */
	protected function initialize() {
		$this->properties               = [];
		$this->class_target_package     = [];
		$this->namespace_target_package = [];
		$this->alternative_instances    = [];
		$this->class_map                = [];
		foreach ( $this->app->get_packages() as $package ) {
			$map = $package->get_config( 'map', $this->app );
			foreach ( $map as $name => $class ) {
				if ( is_array( $class ) ) {
					foreach ( $class as $k => $v ) {
						$class                                = ltrim( $v, '\\' );
						$this->properties[ $k ]               = $class;
						$this->class_target_package[ $class ] = $name;
						if ( ! $this->app->is_valid_package( $name ) ) {
							$this->alternative_instances[ $class ] = $this->get_alternative_instance( $name );
						}
					}
				} else {
					$class                                = ltrim( $class, '\\' );
					$this->properties[ $name ]            = $class;
					$this->class_target_package[ $class ] = $package->get_package();
				}
			}
			$map = $package->get_config( 'class_map', $this->app );
			foreach ( $map as $class_map ) {
				foreach ( $class_map as $from => $to ) {
					$from                              = ltrim( $from, '\\' );
					$to                                = ltrim( $to, '\\' );
					$this->class_map [ $from ]         = $to;
					$this->class_target_package[ $to ] = $package->get_package();
				}
			}
			$this->namespace_target_package[ $package->get_namespace() ] = $package->get_package();
		}
	}

	/**
	 * @param string $package
	 *
	 * @return WP_Framework|null
	 */
	private function get_alternative_instance( $package ) {
		foreach ( $this->app->get_instances() as $instance ) {
			if ( $instance->is_valid_package( $package ) ) {
				return $instance;
			}
		}

		return null;
	}

	/**
	 * @param string $name
	 *
	 * @return Singleton
	 */
	public function get( $name ) {
		if ( isset( $this->properties[ $name ] ) ) {
			$class = $this->properties[ $name ];
			if ( ! isset( $this->property_instances[ $class ] ) ) {
				/** @var Singleton $class */
				try {
					$this->property_instances[ $class ] = $class::get_instance( $this->app );
				} catch ( Exception $e ) {
					WP_Framework::kill( $e->getMessage(), __FILE__, __LINE__ );
				}
			}

			return $this->property_instances[ $class ];
		}
		WP_Framework::kill( $name . ' is undefined.', __FILE__, __LINE__ );

		return null;
	}

	/**
	 * @param string $class
	 *
	 * @return bool
	 */
	public function load_class( $class ) {
		$matches = null;
		$class   = ltrim( $class, '\\' );
		if ( isset( $this->property_instances[ $this->properties['define'] ] ) && preg_match( "#\A{$this->define->plugin_namespace}(.+)\z#", $class, $matches ) ) {
			$class = $matches[1];
			$class = ltrim( $class, '\\' );
			$class = strtolower( $class );
			$path  = $this->define->plugin_src_dir . DS . str_replace( '\\', DS, $class ) . '.php';
			if ( is_readable( $path ) ) {
				/** @noinspection PhpIncludeInspection */
				require_once $path;

				return true;
			}
		} elseif ( isset( $this->class_target_package[ $class ] ) ) {
			if ( array_key_exists( $class, $this->alternative_instances ) ) {
				if ( ! isset( $this->alternative_instances[ $class ] ) ) {
					$this->alternative_instances[ $class ] = $this->get_alternative_instance( $this->class_target_package[ $class ] );
					if ( ! isset( $this->alternative_instances[ $class ] ) ) {
						$this->alternative_instances[ $class ] = $this->app;
					}
				}
				$instance = $this->alternative_instances[ $class ];
			} else {
				$instance = $this->app;
			}
			if ( $instance->get_package_instance( $this->class_target_package[ $class ] )->load_class( $class ) ) {
				return true;
			}
		} elseif ( preg_match( '#\A(\w+)\\\\#', $class, $matches ) && isset( $this->namespace_target_package[ $matches[1] ] ) ) {
			if ( $this->app->get_package_instance( $this->namespace_target_package[ $matches[1] ] )->load_class( $class ) ) {
				return true;
			}
		}

		return false;
	}

	/**
	 * main init
	 */
	public function main_init() {
		if ( $this->initialized ) {
			return;
		}
		$this->initialized = true;

		$this->filter->do_action( 'app_initialize', $this );
		$this->system;
	}

	/**
	 * @return bool
	 */
	public function has_initialized() {
		return $this->initialized;
	}

	/**
	 * @param string $class
	 *
	 * @return array
	 */
	public function get_mapped_class( $class ) {
		return isset( $this->class_map[ $class ] ) ? [ true, $this->class_map[ $class ] ] : [ false, $class ];
	}

	/**
	 * @param string $name
	 *
	 * @return bool
	 */
	public function is_loaded( $name ) {
		return isset( $this->property_instances[ $this->app->plugin_name ][ $name ] );
	}

	/**
	 * @return string
	 * @deprecated
	 */
	public function get_plugin_uri() {
		return $this->app->get_plugin_data( $this->app->is_theme ? 'ThemeURI' : 'PluginURI' );
	}

	/**
	 * @return array
	 * @deprecated
	 */
	public function get_package_versions() {
		return $this->app->array->combine( array_map( function ( $package ) {
			/** @var Package_Base $package */
			return [
				'version' => $package->get_version(),
				'package' => $package->get_package(),
			];
		}, $this->app->get_packages() ), 'package', 'version' );
	}

	/**
	 * @param string $name
	 * @param string|null $key
	 * @param mixed $default
	 *
	 * @return mixed
	 */
	public function get_config( $name, $key = null, $default = null ) {
		return $this->config->get( $name, $key, $default );
	}

	/**
	 * @param string $key
	 * @param mixed $default
	 *
	 * @return mixed
	 */
	public function get_option( $key, $default = '' ) {
		return $this->option->get( $key, $default );
	}

	/**
	 * @param string $key
	 * @param mixed $default
	 *
	 * @return mixed
	 */
	public function get_session( $key, $default = null ) {
		return $this->session->get( $key, $default );
	}

	/**
	 * @param string $key
	 * @param mixed $value
	 * @param int|null $duration
	 */
	public function set_session( $key, $value, $duration = null ) {
		$this->session->set( $key, $value, $duration );
	}

	/**
	 * @param null|string|false $capability
	 *
	 * @return bool
	 */
	public function user_can( $capability = null ) {
		return $this->user->user_can( $capability );
	}

	/**
	 * @param mixed $message
	 * @param mixed $context
	 * @param string $level
	 */
	public function log( $message, $context = null, $level = '' ) {
		if ( ! $this->app->is_valid_package( 'log' ) ) {
			$this->error_log( $message, $context );

			return;
		}
		if ( $message instanceof Exception ) {
			$this->log->log( $message->getMessage(), isset( $context ) ? $context : $message->getTraceAsString(), empty( $level ) ? 'error' : $level );
		} elseif ( $message instanceof WP_Error ) {
			$this->log->log( $message->get_error_message(), isset( $context ) ? $context : $message->get_error_data(), empty( $level ) ? 'error' : $level );
		} else {
			$this->log->log( $message, $context, $level );
		}
	}

	/**
	 * @param mixed $message
	 * @param mixed $context
	 */
	private function error_log( $message, $context ) {
		if ( $message instanceof Exception ) {
			error_log( $message->getMessage() ); // phpcs:ignore WordPress.PHP.DevelopmentFunctions.error_log_error_log
			error_log( print_r( isset( $context ) ? $context : $message->getTraceAsString(), true ) ); // phpcs:ignore WordPress.PHP.DevelopmentFunctions.error_log_error_log, WordPress.PHP.DevelopmentFunctions.error_log_print_r
		} elseif ( $message instanceof WP_Error ) {
			error_log( $message->get_error_message() ); // phpcs:ignore WordPress.PHP.DevelopmentFunctions.error_log_error_log
			error_log( print_r( isset( $context ) ? $context : $message->get_error_data(), true ) ); // phpcs:ignore WordPress.PHP.DevelopmentFunctions.error_log_error_log, WordPress.PHP.DevelopmentFunctions.error_log_print_r
		}
	}

	/**
	 * @param string $message
	 * @param string $group
	 * @param bool $error
	 * @param bool $escape
	 * @param null|array $override_allowed_html
	 */
	public function add_message( $message, $group = '', $error = false, $escape = true, $override_allowed_html = null ) {
		if ( ! $this->app->is_valid_package( 'admin' ) ) {
			return;
		}
		$this->admin->add_message( $message, $group, $error, $escape, $override_allowed_html );
	}

	/**
	 * @param string $send_to
	 * @param string $subject
	 * @param string|array $body
	 * @param string|false $text
	 *
	 * @return bool
	 */
	public function send_mail( $send_to, $subject, $body, $text = false ) {
		if ( ! $this->app->is_valid_package( 'mail' ) ) {
			return false;
		}

		return $this->mail->send( $send_to, $subject, $body, $text );
	}

	/**
	 * @param Package $instance
	 * @param string $name
	 * @param array $args
	 * @param bool $echo
	 * @param bool $error
	 * @param bool $remove_nl
	 *
	 * @return string
	 */
	public function get_view( Package $instance, $name, array $args = [], $echo = false, $error = true, $remove_nl = false ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return '';
		}

		$this->drawer->set_package( $instance );

		return $this->drawer->get_view( $name, $args, $echo, $error, $remove_nl );
	}

	/**
	 * @param Package $instance
	 * @param string $name
	 * @param array $args
	 * @param int $priority
	 */
	public function add_script_view( Package $instance, $name, array $args = [], $priority = 10 ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return;
		}

		$this->drawer->set_package( $instance );
		$this->drawer->add_script_view( $name, $args, $priority );
	}

	/**
	 * @param Package $instance
	 * @param string $name
	 * @param array $args
	 * @param int $priority
	 */
	public function add_style_view( Package $instance, $name, array $args = [], $priority = 10 ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return;
		}

		$this->drawer->set_package( $instance );
		$this->drawer->add_style_view( $name, $args, $priority );
	}

	/**
	 * @param Package $instance
	 * @param string $handle
	 * @param string $file
	 * @param array $depends
	 * @param string|bool|null $ver
	 * @param string $media
	 * @param string $dir
	 */
	public function enqueue_style( Package $instance, $handle, $file, array $depends = [], $ver = false, $media = 'all', $dir = 'css' ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return;
		}

		$this->drawer->set_package( $instance );
		$this->drawer->enqueue_style( $handle, $file, $depends, $ver, $media, $dir );
	}

	/**
	 * @param Package $instance
	 * @param string $handle
	 * @param string $file
	 * @param array $depends
	 * @param string|bool|null $ver
	 * @param bool $in_footer
	 * @param string $dir
	 */
	public function enqueue_script( Package $instance, $handle, $file, array $depends = [], $ver = false, $in_footer = true, $dir = 'js' ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return;
		}

		$this->drawer->set_package( $instance );
		$this->drawer->enqueue_script( $handle, $file, $depends, $ver, $in_footer, $dir );
	}

	/**
	 * @param Package $instance
	 * @param string $handle
	 * @param string $name
	 * @param array $data
	 *
	 * @return bool
	 */
	public function localize_script( Package $instance, $handle, $name, array $data ) {
		if ( ! $this->app->is_valid_package( 'presenter' ) ) {
			return false;
		}

		$this->drawer->set_package( $instance );

		return $this->drawer->localize_script( $handle, $name, $data );
	}

	/**
	 * @param string $file
	 *
	 * @return string
	 */
	public function get_page_slug( $file ) {
		return basename( $file, '.php' );
	}

	/**
	 * @param string $key
	 * @param string|null $target
	 *
	 * @return mixed
	 * @SuppressWarnings(PHPMD.UndefinedVariable)
	 */
	public function get_shared_object( $key, $target = null ) {
		if ( ! isset( $target ) ) {
			$target = $this->app->plugin_name;
		}

		return isset( self::$shared_object[ $target ][ $key ] ) ? self::$shared_object[ $target ][ $key ] : null;
	}

	/**
	 * @param string $key
	 * @param mixed $object
	 * @param string|null $target
	 * @SuppressWarnings(PHPMD.UndefinedVariable)
	 */
	public function set_shared_object( $key, $object, $target = null ) {
		if ( ! isset( $target ) ) {
			$target = $this->app->plugin_name;
		}
		self::$shared_object[ $target ][ $key ] = $object;
	}

	/**
	 * @param string $key
	 * @param string|null $target
	 *
	 * @return bool
	 * @SuppressWarnings(PHPMD.UndefinedVariable)
	 */
	public function isset_shared_object( $key, $target = null ) {
		if ( ! isset( $target ) ) {
			$target = $this->app->plugin_name;
		}

		return isset( self::$shared_object[ $target ] ) && array_key_exists( $key, self::$shared_object[ $target ] );
	}

	/**
	 * @param string $key
	 * @param string|null $target
	 * @SuppressWarnings(PHPMD.UndefinedVariable)
	 */
	public function delete_shared_object( $key, $target = null ) {
		if ( ! isset( $target ) ) {
			$target = $this->app->plugin_name;
		}
		unset( self::$shared_object[ $target ][ $key ] );
	}

	/**
	 * @param string $name
	 * @param callable $func
	 * @param int $timeout
	 *
	 * @return bool
	 */
	public function lock_process( $name, callable $func, $timeout = 60 ) {
		return $this->utility->lock_process( $this->app, $name, $func, $timeout );
	}

	/**
	 * @return bool
	 */
	public function is_enough_version() {
		if ( ! isset( $this->property_instances[ $this->properties['system'] ] ) ) {
			return true;
		}

		return $this->system->is_enough_version();
	}

	/**
	 * load all packages
	 */
	public function load_all_packages() {
		foreach ( array_keys( $this->properties ) as $name ) {
			if ( ! $this->app->is_valid_package( $name ) ) {
				continue;
			}
			$this->$name;
		}
	}
}
